-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

use work.sha3_blake_package.ALL;
use work.sha3_pkg.all;

entity blake_datapath is	 	
	generic (	  	   
		w : integer := 32;
		HS : integer := 256);
	port (
		-- external		
		clk : in std_logic;		
		din  : in std_logic_vector(w-1 downto 0);	
		dout : out std_logic_vector(w-1 downto 0);			
		
		-- internal
		round : in std_logic_vector(7 downto 0);				
		msg_wr, state_wr, h_wr : in std_logic;
		sf : in std_logic;
		
		load_c, ec, eth, etl, dt, smux : in std_logic;
		stbot : in std_logic;
		zc0, zcRemainder, extra_block : out std_logic		
	);				  
end blake_datapath;

architecture struct of blake_datapath is   
	-- CONSTANTS
	constant b : integer := get_b( HS );
	constant log2b			: integer := log2( b );	  
	constant log2bzeros : std_logic_vector(log2b-1 downto 0) := ( others => '0' );
	
	
	-- TYPES
	type mem_type is array ( 0 to 15 ) of std_logic_vector(w-1 downto 0);
	type permute_type is array (0 to 255) of integer;
	type iv_type is array (0 to 1, 0 to 3 ) of std_logic_vector(w-1 downto 0);
	type state_reg_type is array(0 to 3, 0 to 3) of std_logic_vector(w-1 downto 0);
	type state_addr_type is array ( 0 to 7 ) of integer;										 
	type hram_type is array (0 to 1, 0 to 3) of std_logic_vector(w-1 downto 0);
	
	-- FUNCTIONS
	function get_cons_folded ( h : integer; b : integer ; iw : integer) return mem_type is
		variable const : std_logic_vector(b-1 downto 0);
		variable ret : mem_type;
	begin
		if ( h <= 256 ) then
			const := switch_endian_word(const32,b,iw);
		else 
			const := switch_endian_word(const64,b,iw);
		end if;
		for i in 0 to 15 loop
			ret(15-i) := const(iw*(i+1) - 1 downto iw*i);
		end loop;
		return ret;
	end function get_cons_folded;	  

	function get_const_addr_lookup ( msg_addr_lookup : permute_type ) return permute_type is 
		variable ret : permute_type;
	begin
		for i in 0 to 127 loop
			ret(2*i)   := msg_addr_lookup(2*i+1);
			ret(2*i+1) := msg_addr_lookup(2*i);			
		end loop;
		return ret;
	end function get_const_addr_lookup;
	
	function get_iv_folded ( h : integer; iv : std_logic_vector ) return iv_type is
		variable iw : integer := get_iw( h );
		variable b : integer := get_b( h ) ;
		variable ret : iv_type;
	begin
		for i in 0 to 3 loop
			ret(1,3-i) := iv(iw*(i+1) - 1 downto iw*i);
			ret(0,3-i) := iv(iw*(i+1)+b/4 - 1 downto iw*i+b/4);
		end loop;
		return ret;
	end function get_iv_folded;
	
		-- create another constant rom so that it doesn't use up another constant table
	function get_constant2 ( h : integer ) return iv_type is	
		variable ret : iv_type;
	begin
		if ( h = 256 ) then
			ret := ((x"243F6A88" , x"85A308D3" , x"13198A2E" , x"03707344" ), 
					(x"A4093822" , x"299F31D0" , x"082EFA98" , x"EC4E6C89" )); 		  
		else
			ret := ((x"243F6A8885A308D3" , x"13198A2E03707344" , x"A4093822299F31D0" , x"082EFA98EC4E6C89") ,
					(x"452821E638D01377" , x"BE5466CF34E90C6C" , x"C0AC29B7C97C50DD" , x"3F84D5B5B5470917"));
		end if;
		return ret;
	end function get_constant2;
		
		
	-- MORE CONSTANTS
	constant cons_init : iv_type := get_constant2( HS );
	constant cons_rom : mem_type := get_cons_folded(HS,b,w);	
	constant msg_addr_lookup    : permute_type :=  ( 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15,
													 14, 10, 4, 8, 9, 15, 13, 6, 1, 12, 0, 2, 11, 7, 5, 3,
													 11, 8, 12, 0, 5, 2, 15, 13, 10, 14, 3, 6, 7, 1, 9, 4,
													 7, 9, 3, 1, 13, 12, 11, 14, 2, 6, 5, 10, 4, 0, 15, 8,
													 9, 0, 5, 7, 2, 4, 10, 15, 14, 1, 11, 12, 6, 8, 3, 13,
													 2, 12, 6, 10, 0, 11, 8, 3, 4, 13, 7, 5, 15, 14, 1, 9,
													 12, 5, 1, 15, 14, 13, 4, 10, 0, 7, 6, 3, 9, 2, 8, 11,
													 13, 11, 7, 14, 12, 1, 3, 9, 5, 0, 15, 4, 8, 6, 2, 10,
													 6, 15, 14, 9, 11, 3, 0, 8, 12, 2, 13, 7, 1, 4, 10, 5,
													 10, 2, 8, 4, 7, 6, 1, 5, 15, 11, 9, 14, 3, 12, 13, 0,
													 -- loop back
													 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15,
													 14, 10, 4, 8, 9, 15, 13, 6, 1, 12, 0, 2, 11, 7, 5, 3,
													 11, 8, 12, 0, 5, 2, 15, 13, 10, 14, 3, 6, 7, 1, 9, 4,
													 7, 9, 3, 1, 13, 12, 11, 14, 2, 6, 5, 10, 4, 0, 15, 8,
													 9, 0, 5, 7, 2, 4, 10, 15, 14, 1, 11, 12, 6, 8, 3, 13,
													 2, 12, 6, 10, 0, 11, 8, 3, 4, 13, 7, 5, 15, 14, 1, 9);	
	constant const_addr_lookup    : permute_type :=  get_const_addr_lookup( msg_addr_lookup );													 
    constant iv_temp : std_logic_vector(b/2-1 downto 0) := get_iv( HS, w );
	constant iv : iv_type := get_iv_folded( HS, iv_temp );
	
	constant state_addr1_lookup : state_addr_type := (0, 1, 2, 3, 1, 2, 3, 0); 
	constant state_addr2_lookup : state_addr_type := (0, 1, 2, 3, 2, 3, 0, 1);
	constant state_addr3_lookup : state_addr_type := (0, 1, 2, 3, 3, 0, 1, 2);
	-- SIGNALS
	
	-- CONTROL
		signal c_s : std_logic_vector(15 downto 0);
		signal t : std_logic_vector(2*w-1 downto 0);
		signal t_top : std_logic_vector(2*w-log2b-1 downto 0);
		signal t_bot, t_bot_enabled : std_logic_vector(log2b-1 downto 0);
		
		-- PERMUTE
		signal msg_raddr, const_raddr : std_logic_vector(3 downto 0);
		signal msg : std_logic_vector(w-1 downto 0);
	    signal cm, const : std_logic_vector(w-1 downto 0);
		-- H
		signal hin_p4, hin_p0, hout_p4, hout_p0 : std_logic_vector(w-1 downto 0);
		-- INIT
		signal iv_addr : std_logic_vector(1 downto 0);
		signal chain_p0, chain_p4 : std_logic_vector(w-1 downto 0);
		signal init_o_p0, init_o_p4, init_o_p8, init_o_p12, t_sel : std_logic_vector(w-1 downto 0);
		
		-- CORE								   
		signal state_in_p0, state_in_p4, state_in_p8, state_in_p12 : std_logic_vector(w-1 downto 0);
		signal gin0, gin1, gin2, gin3, gout0, gout1, gout2, gout3 : std_logic_vector(w-1 downto 0);
		
		type sram_addr_type is array ( 0 to 3 ) of std_logic_vector(1 downto 0);
		signal sram_addr, sram_addr_delayed : sram_addr_type;
		--signal sram_addr, sram_addr1, sram_addr2, sram_addr3 : std_logic_vector(1 downto 0);
		signal sram_addr_lookup : std_logic_vector(2 downto 0);
		-- RAM SIGNALS				 
		signal msg_ram : mem_type;	  
		signal hram : hram_type;	  
		signal state_ram : state_reg_type;	
		
		signal round_delayed : std_logic_vector(2 downto 0);
begin		
	-- t gens		
	decounter_gen : decountern generic map ( N => 16, sub => 1 ) port map ( clk => clk, rst => '0', load => load_c, en => ec, input => din(log2b+15 downto log2b), output => c_s );
	th_gen : countern generic map ( N => 2*w-log2b, step => 1 ) port map ( clk => clk, rst => dt, load => '0', en => eth, input => conv_std_logic_vector(0,2*w-log2b), output => t_top );
	tl_gen : regn generic map ( N => log2b, init => log2bzeros ) port map ( clk => clk, rst => dt, en => etl, input => din(log2b-1 downto 0), output => t_bot );
	t_bot_enabled <= t_bot when (stbot = '1') else (others => '0');
	t <= t_top & t_bot_enabled;
	zc0 <= '1' when (c_s = 0)  else '0'; 	
	zcRemainder <= '1' when (t_bot /= 0) else '0';
	extra_block <= '1' when (t_bot > b-66) else '0';
	
	
	-- /////////////////////////////
	-- Permute unit
	-- Message Ram and constant rom
	msg_raddr <= conv_std_logic_vector(msg_addr_lookup( conv_integer(round) ), 4 );
	const_raddr <= conv_std_logic_vector(const_addr_lookup( conv_integer(round) ), 4 );
	
	mram_crom_cm: process( clk )	
	begin
		if rising_edge( clk ) then
			if msg_wr = '1' then
				msg_ram(conv_integer(msg_raddr)) <= din;
			end if;
			-- cm <= const xor msg;
		end if;	
	end process;	   
	msg <= msg_ram(conv_integer(msg_raddr));
	const <= cons_rom(conv_integer(const_raddr));
	
	
	
	reg: process( clk ) 
	begin
		if rising_edge( clk ) then
			round_delayed <= round(2 downto 0);
			cm <= const xor msg;
		end if;
	end process;
	-- /////////////////////////////
	-- Initialization unit
	
	iv_addr <=  round_delayed(1 downto 0);
	hin_p0 <= iv(0,conv_integer(iv_addr)) when sf = '1' else chain_p0;
	hin_p4 <= iv(1,conv_integer(iv_addr)) when sf = '1' else chain_p4;
		
	t_sel <= t(w-1 downto 0) when round_delayed(1) = '0' else t(2*w-1 downto w);	
			 
	init_o_p0 <= hin_p0;
	init_o_p4 <= hin_p4;
	init_o_p8 <= cons_init(0,conv_integer(iv_addr));
	init_o_p12 <= cons_init(1,conv_integer(iv_addr)) xor t_sel;
	
	-- /////////////////////////////
	-- H rams
	hreg_gen : process ( clk )
	begin
		if rising_edge( clk ) then
			if h_wr = '1' then
				hram(0,conv_integer(iv_addr)) <= hin_p0;
				hram(1,conv_integer(iv_addr)) <= hin_p4;
			end if;
		end if;		
	end process;	
	hout_p0 <= hram(0,conv_integer(iv_addr));
	hout_p4 <= hram(1,conv_integer(iv_addr));
	
	-- /////////////////////////////
	-- CORE
	
		
		-- Memory 
		
		sram_addr_lookup <= round(3 downto 1) when smux = '0' else '0' & round(1 downto 0);
		sram_addr(0) <= round(2 downto 1) when smux = '0' else round(1 downto 0);
		sram_addr(1) <= conv_std_logic_vector(state_addr1_lookup(conv_integer( sram_addr_lookup ) ), 2);
		sram_addr(2) <= conv_std_logic_vector(state_addr2_lookup(conv_integer( sram_addr_lookup ) ), 2);
		sram_addr(3) <= conv_std_logic_vector(state_addr3_lookup(conv_integer( sram_addr_lookup ) ), 2);
				
		state_reg_gen : process ( clk )		
		begin
			if rising_edge( clk ) then
				if state_wr = '1' then
					state_ram(0,conv_integer(sram_addr_delayed(0))) <= state_in_p0;
					state_ram(1,conv_integer(sram_addr_delayed(1))) <= state_in_p4;
					state_ram(2,conv_integer(sram_addr_delayed(2))) <= state_in_p8;
					state_ram(3,conv_integer(sram_addr_delayed(3))) <= state_in_p12;
				end if;													
				sram_addr_delayed <= sram_addr;									 				
			end if;				
		end process;
		gin0 <= state_ram(0,conv_integer(sram_addr_delayed(0)));
		gin1 <= state_ram(1,conv_integer(sram_addr_delayed(1)));
		gin2 <= state_ram(2,conv_integer(sram_addr_delayed(2)));
		gin3 <= state_ram(3,conv_integer(sram_addr_delayed(3)));		
		
		-- Gfunc half				   
		gfunc_gen : entity work.gfunc_half_modified(struct) 	generic map ( iw => w, h => HS  )
							port map ( ain => gin0, bin => gin1,cin => gin2,din => gin3,
								  const => cm, sel => round_delayed(0),
								  aout => gout0, bout => gout1, cout=> gout2,dout => gout3);
	
		-- Muxes
		state_in_p0 <= init_o_p0 when h_wr = '1' else gout0;
		state_in_p4 <= init_o_p4 when h_wr = '1' else gout1;
		state_in_p8 <= init_o_p8 when h_wr = '1' else gout2;
		state_in_p12 <= init_o_p12 when h_wr = '1' else gout3;
		
	-- /////////////////////
	-- Finalization	
	chain_p0 <= (gin0 xor gin2) xor hout_p0;
	chain_p4 <= (gin1 xor gin3) xor hout_p4;
	
	dout <= chain_p4 when round_delayed(2) = '1' else chain_p0;
	
end struct;